/*
 * Copyright (c) 2007 Anthony Mallet                  --  Wed Feb 21 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdio.h>
#include <err.h>

#include "viam_private.h"

#ifdef VIAM_OPENCV_SUPPORT
# ifndef CV_INLINE
#  define CV_INLINE	static inline
# endif
# define CV_NO_BACKWARD_COMPATIBILITY
# include <cxtypes.h>
# include <cv.h>
#endif


/* --- viam_color_geosetup ------------------------------------------- */

/** \brief Geometry for color filters that do not change geometry.
 *
 * \return 0 on success.
 */

int
viam_color_geosetup(viam_handle_t vh, viam_image_t image,
		    viam_filterargs_t *args, viam_geometry_t *geom, void **data)
{
  /* check validity */
  if (geom->channels != 3) {
    warnx("color filters require three channels image");
    return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }

  /* geometry is unchanged */
  return 0;
}


/* --- viam_bayer_geosetup ------------------------------------------- */

/** \brief Compute geometry of bayer filter.
 *
 * \return 0 on success.
 */

#ifdef VIAM_OPENCV_SUPPORT

int
viam_bayer_geosetup(viam_handle_t vh, viam_image_t image,
		    viam_filterargs_t *args, viam_geometry_t *geom, void **data)
{
  /* check validity */
  if (geom->channels != 1) {
    warnx("bayer requires single channels image");
    return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }

  /* update channels */
  geom->channels = 3;

  return 0;
}

#endif /* VIAM_OPENCV_SUPPORT */

/* --- viam_bayer ---------------------------------------------------- */

/** \brief Transform image with Bayer mosaic.
 *
 * \param[in] vh	Library handle.
 *
 * \return 0 on success.
 */

#ifdef VIAM_OPENCV_SUPPORT

int
viam_bayer(viam_handle_t vh, viam_image_t image, const IplImage *source,
	   IplImage *result, viam_filterargs_t *args, void *data)
{
  int code;

  switch(args->type) {
    case VIAM_FILTER_BAYER_BGGR:	code = CV_BayerBG2BGR;	break;
    case VIAM_FILTER_BAYER_GBRG:	code = CV_BayerGB2BGR;	break;
    case VIAM_FILTER_BAYER_RGGB:	code = CV_BayerRG2BGR;	break;
    case VIAM_FILTER_BAYER_GRBG:	code = CV_BayerGR2BGR;	break;

    default:
      return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }
  cvCvtColor(source, result, code);
  if (cvGetErrStatus()) return viam_seterrno(vh, Eviamlib_OPENCV);

  return 0;
}

#endif /* VIAM_OPENCV_SUPPORT */
