/*
 * Copyright (c) 2007 LAAS/CNRS                       --  Wed Jan 24 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <err.h>

#include "viam_private.h"

#ifdef VIAM_OPENCV_SUPPORT
# ifndef CV_INLINE
#  define CV_INLINE	static inline
# endif
# define CV_NO_BACKWARD_COMPATIBILITY
# include <cxtypes.h>
# include <cv.h>
#endif


/* --- local data ---------------------------------------------------- */

typedef enum viam_filterdst_t {
  VIAM_FDST_SOURCE,
  VIAM_FDST_SWAP,
  VIAM_FDST_FINAL
} viam_filterdst_t;

struct viam_filter_t {
  int refs;
  viam_filtertype_t type;
  viam_filterargs_t args;
};

struct viam_filterinstance_t {
  struct viam_filterinstance_t *next;
  struct viam_filterinstance_t *prec;
  viam_filter_t filter;
  viam_filtermethod_t method;	/**< filter method */
  viam_filterautomode_t autoarg;/**< (man, auto, one-push) */
  viam_filterargs_t values;	/**< current filter values */
  viam_filterdst_t dst;		/**< destination buffer (swap or final) */

  IplImage *source;		/**< pointer to source image */
  IplImage result;		/**< image header of the filtered source */
  viam_geometry_t geom;		/**< resulting geometry */

  void *data;			/**< filter-specific data */
};

static const struct viam_filterfunc_t {
  const char *name;
  int (*setup)(viam_handle_t vh, viam_image_t camera, viam_filterargs_t *args,
	viam_geometry_t *geom, void **data);
  int (*filter)(viam_handle_t vh, viam_image_t camera, const IplImage *source,
	IplImage *result, viam_filterargs_t *args, void *data);
  int (*autoarg)(viam_handle_t vh, viam_image_t camera, const IplImage *source,
	viam_filterargs_t *args, void *data);
} filter_funcs[] = {
  [VIAM_FILTER_SUBSAMPLE] = {
    "subsampling", viam_geo_geosetup, viam_geo_subsample, NULL
  },
#ifdef VIAM_OPENCV_SUPPORT
  [VIAM_FILTER_DISTORT] = {
    "distortion compensation", viam_geo_distortsetup, viam_geo_remap, NULL
  },
  [VIAM_FILTER_RECTIFY] = {
    "rectification", viam_geo_rectifysetup, viam_geo_remap, NULL
  },
  [VIAM_FILTER_WHITE_BALANCE] = {
    "white balance", viam_color_geosetup, viam_awb, viam_awb_auto
  },
  [VIAM_FILTER_BAYER_BGGR] = {
    "Bayer BGGR", viam_bayer_geosetup, viam_bayer, NULL
  },
  [VIAM_FILTER_BAYER_GRBG] = {
    "Bayer GRBG", viam_bayer_geosetup, viam_bayer, NULL
  },
  [VIAM_FILTER_BAYER_RGGB] = {
    "Bayer RGGB", viam_bayer_geosetup, viam_bayer, NULL
  },
  [VIAM_FILTER_BAYER_GBRG] = {
    "Bayer GBRG", viam_bayer_geosetup, viam_bayer, NULL
  }
#endif /* VIAM_OPENCV_SUPPORT */
};

static viam_filter_t	viam_filter_create(viam_handle_t vh);
static const char *	viam_filter_methoddescr(viam_filtermethod_t method);
static const char *	viam_filter_autodescr(viam_filterautomode_t aumode);


/* --- viam_filter_image --------------------------------------------- */

/** \brief Iterate over filters and apply them on image.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Image object.
 * \return 0 on success.
 */

int
viam_filter_image(viam_handle_t vh, viam_image_t image, IplImage *final)
{
  viam_camera_t camera = viam_image_camera(image);
  viam_filterinstance_t f;
  int s;

  /* handle special case with no filter separately */
  if (!viam_image_fi(image)) {
    memcpy(
      final->imageData, viam_camera_raw(camera)->imageData, final->imageSize);
    return 0;
  }

  /* interate over all filters for the general case */
  for(f = viam_image_fi(image); f; f=f->next) {
    if (f->autoarg == VIAM_FILTER_SOFTWARE_AUTO) {
      s = filter_funcs[f->filter->type].autoarg(
	vh, image, f->source, &f->values, f->data);
      if (s) return s;
    }

    switch(f->method) {
      case VIAM_FILTER_SOFTWARE:
	if (f->dst == VIAM_FDST_SWAP)
	  cvSetData(&f->result, viam_swap(), CV_AUTOSTEP);
	else
	  cvSetData(&f->result, final->imageData, final->widthStep);

	s = filter_funcs[f->filter->type].filter(
	  vh, image, f->source, &f->result, &f->values, f->data);
	if (s) return s;
	break;

      case VIAM_FILTER_HARDWARE:
	cvSetData(&f->result, f->source->imageData, f->source->widthStep);
	if (f->autoarg == VIAM_FILTER_SOFTWARE_AUTO) {
	  s = viam_sethwfeature(vh, camera, f->filter->type,
				f->method, f->autoarg, &f->values);
	  if (s) return s;
	}
	break;

      case VIAM_FILTER_INVALID: break;
      case VIAM_FILTER_OFF: break;
    }
  }

  return 0;
}


/* --- viam_filter_push ---------------------------------------------- */

/** \brief Install filter on an image.
 *
 * The camera must not be configured.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera object.
 * \param[in] filter	Filter object.
 * \param[in] method	Filter method.
 * \param[in] aumode	Auto mode.
 *
 * \return 0 on success.
 */

int
viam_filter_push(viam_handle_t vh,
		 viam_image_t image, viam_filter_t filter,
		 viam_filtermethod_t method, viam_filterautomode_t aumode)
{
  viam_filterinstance_t prec, fi;
  viam_camera_t camera = viam_image_camera(image);
  int s;

  if (viam_camera_setup(camera))
    return viam_seterrno(vh, Eviamlib_CONFIGURED);

  /* check mode has only one bit set (i.e. exactly one mode) */
  if (method & (method-1)) {
    warnx("bad method %d", method);
    return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }
  if (aumode & (aumode-1)) {
    warnx("bad auto mode %d", aumode);
    return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }

  /* find previous filter */
  for(prec = viam_image_fi(image); prec && prec->next; prec = prec->next)
    /* empty body */;

  /* make new filter instance structure */
  fi = malloc(sizeof(*fi));
  if (!fi) {
    warnx("unable to allocate memory");
    return viam_seterrno(vh, Eviamlib_MEMORY);
  }
  fi->next = NULL;
  fi->prec = prec;
  filter->refs++;
  fi->filter = filter;
  fi->method = method;
  fi->autoarg = aumode;
  fi->values = filter->args;
  fi->dst = VIAM_FDST_FINAL; /* since we are the last filter */
  fi->source = prec ? &prec->result : viam_camera_raw(camera);
  fi->geom.width = fi->source->width;
  fi->geom.height = fi->source->height;
  fi->geom.scalex = prec ? prec->geom.scalex : 1.;
  fi->geom.scaley = prec ? prec->geom.scaley : 1.;
  fi->geom.offx = prec ? prec->geom.offx : 0.;
  fi->geom.offy = prec ? prec->geom.offy : 0.;
  fi->geom.depth = fi->source->depth;
  fi->geom.channels = fi->source->nChannels;
  fi->data = NULL;

  /* configure geometry and modes */
  s = viam_filter_configure(vh, image, fi);
  if (s) return viam_errno(vh);

  /* insert new filter at tail */
  if (prec) prec->next = fi; else viam_image_setfi(image, fi);

  warnx("added %s filter to image %s",
	filter_funcs[filter->type].name, viam_image_name(image));
  return 0;
}


/* --- viam_filter_configure ----------------------------------------- */

/** \brief (Re)configure images geometry and filter mode.
 *
 * \param[in] vh	Library handle.
 * \param[in] image	Image object.
 * \param[in] fi	Filter instances.
 *
 * \return 0 on success.
 */

int
viam_filter_configure(viam_handle_t vh, viam_image_t image,
		      viam_filterinstance_t fi)
{
  viam_camera_t camera = viam_image_camera(image);
  viam_filterinstance_t f;
  unsigned int methods, aumodes;
  viam_filterdst_t dst;
  viam_geometry_t *pgeo;
  int s, count;

  if (fi && fi->prec) pgeo = &fi->prec->geom; else pgeo = NULL;
  for(f = fi; f; f = f->next) {
    /* if camera is not configured or has no hw mode, stop here */
    if (!viam_camera_hwdata(camera) || !viam_camera_hwmode(vh, camera))
      continue;

    /* check geometry */
    if (!filter_funcs[f->filter->type].setup)
      return viam_seterrno(vh, Eviamlib_NOT_IMPLEMENTED);

    f->geom.width = f->source->width;
    f->geom.height = f->source->height;
    f->geom.scalex = pgeo ? pgeo->scalex : 1.;
    f->geom.scaley = pgeo ? pgeo->scaley : 1.;
    f->geom.offx = pgeo ? pgeo->offx : 0.;
    f->geom.offy = pgeo ? pgeo->offy : 0.;
    f->geom.depth = f->source->depth;
    f->geom.channels = f->source->nChannels;
    pgeo = &f->geom;

    s = filter_funcs[f->filter->type].setup(
      vh, image, &f->values, &f->geom, &f->data);
    if (s) {
      warnx("%s filter on image %s has invalid parameters",
	    filter_funcs[f->filter->type].name, viam_image_name(image));
      return s;
    }

    /* initialize image header */
    cvInitImageHeader(&f->result, cvSize(f->geom.width, f->geom.height),
		      f->geom.depth, f->geom.channels, IPL_ORIGIN_TL, 4);
    viam_setswap(f->result.imageSize);

    warnx("%s filter on image %s produces %dx%d images, %d bpp, %d channel%s",
	  filter_funcs[f->filter->type].name, viam_image_name(image),
	  f->geom.width, f->geom.height, f->geom.depth,
	  f->geom.channels, f->geom.channels>1?"s":"");

    /* check filter mode */
    s = viam_queryfeature(vh, camera, f->filter->type, &methods, &aumodes);
    if (s) return s;

    if (!(methods & f->method)) {
      warnx("camera %s has no %s %s", viam_camera_name(camera),
	    viam_filter_methoddescr(f->method), filter_funcs[f->filter->type].name);
      return viam_seterrno(vh, Eviamlib_NOT_IMPLEMENTED);
    }
    if (!(aumodes & f->autoarg)) {
      warnx("camera %s has no %s %s", viam_camera_name(camera),
	    viam_filter_methoddescr(f->method), filter_funcs[f->filter->type].name);
      return viam_seterrno(vh, Eviamlib_NOT_IMPLEMENTED);
    }

    /* configure hw filters now */
    if (f->method == VIAM_FILTER_OFF || f->method == VIAM_FILTER_HARDWARE) {
      s = viam_sethwfeature(vh, camera, f->filter->type,
			    f->method, f->autoarg, &f->values);
      if (s) return s;
    }
  }

  /* update images destinations for all filters */

  count = 0;
  for(f = viam_image_fi(image); f; f = f->next)
    if (f->method == VIAM_FILTER_SOFTWARE) count++;

  dst = (count%2) ? VIAM_FDST_FINAL : VIAM_FDST_SWAP;

  for(f = viam_image_fi(image); f; f = f->next) {
    f->dst = dst;
    if (f->method == VIAM_FILTER_SOFTWARE)
      dst = (dst == VIAM_FDST_FINAL) ? VIAM_FDST_SWAP : VIAM_FDST_FINAL;
  }

  return 0;
}


/* --- viam_filter_geo ----------------------------------------------- */

/** \brief Get filtered image geometry.
 *
 * \param[in] vh	Library handle.
 * \param[in] fi	Filter instances.
 * \param[out] width	Width in pixels.
 * \param[out] height	Height in pixels.
 * \param[out] depth	Bit depth of pixels.
 * \param[out] channels	Number of channels per elements.
 *
 * \return 0 on success.
 */

int
viam_filter_geo(viam_handle_t vh, viam_filterinstance_t fi,
		viam_geometry_t *geom)
{
  viam_filterinstance_t f;

  for(f = fi; f && f->next; f = f->next)
    /* empty body */;
  if (!f) return viam_seterrno(vh, Eviamlib_IMAGE_GEOMETRY);

  *geom = f->geom;
  return 0;
}


/* --- viam_filter_queryswfeature ------------------------------------ */

/** \brief Return software filter capabilities.
 *
 * \param[in] vh	Library handle.
 * \param[in] camera	Camera object.
 * \param[in] filter	Filter type.
 * \param[out] methods	Or'ed combination of capabilities.
 * \param[out] aumodes	Or'ed combination of auto modes.
 *
 * \return 0 on success.
 */

int
viam_filter_queryswfeature(viam_handle_t vh, viam_filtertype_t filter,
			   unsigned int *methods, unsigned int *aumodes)
{
  if (methods) *methods = 0;
  if (aumodes) *aumodes = 0;

  if (filter_funcs[filter].filter) {
    if (methods) *methods |= VIAM_FILTER_SOFTWARE;
    if (aumodes) *aumodes |= VIAM_FILTER_MANUAL;
  }

  if (filter_funcs[filter].autoarg && aumodes) {
    *aumodes |= VIAM_FILTER_SOFTWARE_AUTO;
    *aumodes |= VIAM_FILTER_SOFTWARE_ONE_PUSH;
  }

  return 0;
}


/* --- viam_filter_subsample_create ---------------------------------- */

/** \brief Create a subsampling filter.
 *
 * Crop and scale image according to new size specification. If sizes are
 * floats between 0 and 1, they are interpreted as a percentage of the
 * original dimension. Otherwise they are rounded to the nearest
 * integer.
 *
 * \param[in] left	Left corner.
 * \param[in] top	Top corner.
 * \param[in] bottom	Right corner.
 * \param[in] right	Bottom corner.
 * \param[in] towidth	Desired window width.
 * \param[in] toheight	Desired window height.
 */

viam_filter_t
viam_filter_subsample_create(viam_handle_t vh,
			     double left, double top, double right, double bottom,
			     double towidth, double toheight)
{
  viam_filter_t f = viam_filter_create(vh);
  if (!f) return NULL;

  f->type = VIAM_FILTER_SUBSAMPLE;
  f->args.left = left;
  f->args.top = top;
  f->args.right = right;
  f->args.bottom = bottom;
  f->args.towidth = towidth;
  f->args.toheight = toheight;
  return f;
}


/* --- viam_filter_distort_create ------------------------------------ */

/** \brief Create a distortion compensation filter.
 *
 * Crop and scale image according to new size specification. If sizes are
 * floats between 0 and 1, they are interpreted as a percentage of the
 * original dimension. Otherwise they are rounded to the nearest
 * integer. Camera must be calibrated.
 *
 * \param[in] left	Left corner.
 * \param[in] top	Top corner.
 * \param[in] bottom	Right corner.
 * \param[in] right	Bottom corner.
 * \param[in] towidth	Desired window width.
 * \param[in] toheight	Desired window height.
 */

#ifdef VIAM_OPENCV_SUPPORT

viam_filter_t
viam_filter_distort_create(viam_handle_t vh,
			   double left, double top, double right, double bottom,
			   double towidth, double toheight)
{
  viam_filter_t f = viam_filter_create(vh);
  if (!f) return NULL;

  f->type = VIAM_FILTER_DISTORT;
  f->args.left = left;
  f->args.top = top;
  f->args.right = right;
  f->args.bottom = bottom;
  f->args.towidth = towidth;
  f->args.toheight = toheight;
  return f;
}

#endif /* VIAM_OPENCV_SUPPORT */


/* --- viam_filter_rectify_create ------------------------------------ */

/** \brief Create a rectification filter.
 *
 * Crop and scale image according to new size specification. If sizes are
 * floats between 0 and 1, they are interpreted as a percentage of the
 * original dimension. Otherwise they are rounded to the nearest
 * integer. Camera must be calibrated.
 *
 * \param[in] left	Left corner.
 * \param[in] top	Top corner.
 * \param[in] bottom	Right corner.
 * \param[in] right	Bottom corner.
 * \param[in] towidth	Desired window width.
 * \param[in] toheight	Desired window height.
 */

#ifdef VIAM_OPENCV_SUPPORT

viam_filter_t
viam_filter_rectify_create(viam_handle_t vh,
			   double left, double top, double right, double bottom,
			   double towidth, double toheight)
{
  viam_filter_t f = viam_filter_create(vh);
  if (!f) return NULL;

  f->type = VIAM_FILTER_RECTIFY;
  f->args.left = left;
  f->args.top = top;
  f->args.right = right;
  f->args.bottom = bottom;
  f->args.towidth = towidth;
  f->args.toheight = toheight;
  return f;
}

#endif /* VIAM_OPENCV_SUPPORT */


/* --- viam_filter_whitebalance_create ------------------------------- */

/** \brief Create a white balance filter.
 *
 * Adjust gain for U (blue) and V (red) channel.
 *
 * \param[in] vh	Library handle.
 * \param[in] blue	Blue gain% in [0:1].
 * \param[in] red	Red gain% in [0:1].
 *
 * \return initialized filter.
 */

#ifdef VIAM_OPENCV_SUPPORT

viam_filter_t
viam_filter_whitebalance_create(viam_handle_t vh, double blue, double red)
{
  viam_filter_t f = viam_filter_create(vh);
  if (!f) return NULL;

  f->type = VIAM_FILTER_WHITE_BALANCE;
  f->args.blue = blue;
  f->args.red = red;
  return f;
}

#endif /* VIAM_OPENCV_SUPPORT */


/* --- viam_filter_bayer_create -------------------------------------- */

/** \brief Create a bayer filter.
 *
 * Convert image according to the specified bayer pattern.
 *
 * \param[in] vh	Library handle.
 * \param[in] pattern	Filter type: specific bayer patter.
 *
 * \return initialized filter.
 */

#ifdef VIAM_OPENCV_SUPPORT

viam_filter_t
viam_filter_bayer_create(viam_handle_t vh, viam_filtertype_t pattern)
{
  viam_filter_t f;

  switch(pattern) {
    case VIAM_FILTER_BAYER_BGGR:
    case VIAM_FILTER_BAYER_GBRG:
    case VIAM_FILTER_BAYER_RGGB:
    case VIAM_FILTER_BAYER_GRBG:
      break;

    default:
      viam_seterrno(vh, Eviamlib_INVALID_VALUE);
      return NULL;
  }

  f = viam_filter_create(vh);
  if (!f) return NULL;

  f->type = pattern;
  f->args.type = pattern;
  return f;
}

#endif /* VIAM_OPENCV_SUPPORT */


/* --- viam_filter_create -------------------------------------------- */

/** \brief Create a generic filter structure.
 *
 * \return the empty structure.
 */

static viam_filter_t
viam_filter_create(viam_handle_t vh)
{
  viam_filter_t f;

  f = malloc(sizeof(*f));
  if (!f) {
    warnx("unable to allocate memory");
    viam_seterrno(vh, Eviamlib_MEMORY);
    return NULL;
  }

  f->refs = 0;
  return f;
}


/* --- viam_filter_methoddescr --------------------------------------- */

static const char *
viam_filter_methoddescr(viam_filtermethod_t method)
{
  switch(method) {
    case VIAM_FILTER_INVALID:	return "invalid";
    case VIAM_FILTER_OFF:	return "hardware off";
    case VIAM_FILTER_HARDWARE:	return "hardware";
    case VIAM_FILTER_SOFTWARE:	return "software";
  }

  return "*null*";
}


/* --- viam_filter_autoescr ------------------------------------------ */

static const char *
viam_filter_autodescr(viam_filterautomode_t aumode)
{
  switch(aumode) {
    case VIAM_FILTER_AUTOINVALID:	return "invalid";
    case VIAM_FILTER_MANUAL:		return "manual";
    case VIAM_FILTER_HARDWARE_AUTO:	return "hardware auto";
    case VIAM_FILTER_HARDWARE_ONE_PUSH:	return "hardware one push";
    case VIAM_FILTER_SOFTWARE_AUTO:	return "software auto";
    case VIAM_FILTER_SOFTWARE_ONE_PUSH:	return "software one push";
  }

  return "*null*";
}
