/*
 * Copyright (c) 2007 Anthony Mallet                  --  Wed Feb 21 2007
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <stdio.h>
#include <err.h>

#include "viam_private.h"

#ifdef VIAM_OPENCV_SUPPORT
# ifndef CV_INLINE
#  define CV_INLINE	static inline
# endif
# define CV_NO_BACKWARD_COMPATIBILITY
# include <cxtypes.h>
# include <cv.h>
#endif


/* --- viam_geo_subsample -------------------------------------------- */

/** \brief Subsample image.
 *
 * \param[in] vh	Library handle
 * \param[in] image	Image used for initial acquisition
 * \param[in] source	Source image
 * \param[out] result	Resulting image
 * \param[in] args	Filter arguments
 *
 * \return 0 on success.
 */

int
viam_geo_subsample(viam_handle_t vh, viam_image_t image, const IplImage *source,
		   IplImage *result, viam_filterargs_t *args, void *data)
{
#ifdef VIAM_OPENCV_SUPPORT
  CvRect crop;
  int s = 0;

  /* result image has already it's header correctly set by
   * viam_filter_updategeo(). */

  crop.x = args->left;
  crop.y = args->top;
  crop.width = args->right - args->left + 1;
  crop.height = args->bottom - args->top + 1;
  cvSetImageROI((IplImage *)source, crop);
  cvResize(source, result, CV_INTER_NN);
  if (cvGetErrStatus()) s = Eviamlib_OPENCV;
  cvResetImageROI((IplImage *)source);

  return s;
#else
  unsigned char *l, *p, *ol, *op;
  unsigned int sum, w;
  double sx, sy;
  int i, j, oi, oj, c;

  sx = (double)source->width / result->width;
  sy = (double)source->height / result->height;

  for(i=0,l=result->imageData; i<result->height; i++,l+=result->widthStep)
    for(j=0, p=l; j<=result->width; j++,p+=(result->depth>>3)*result->nChannels)
      for(c=0; c<result->nChannels; c++) {
	sum = 0;
	ol = source->imageData + (int)(i*sy) * source->widthStep;
	ol+= (int)(j*sx) * (source->depth>>3)*source->nChannels;
	for(oi=0; oi<sy; oi++, ol+=source->widthStep)
	  for(oj=0, op=ol; oj<sx; oj++, op+=(source->depth>>3)*source->nChannels)
	    switch (result->depth) {
	      case 8:	sum += op[c]; break;
	      case 16:	sum += ((unsigned short *)op)[c]; break;
	    }

	w = ceil(sx)*ceil(sy); if (w < 1) w = 1;
	sum /= w; if (sum > 255) sum = 255;
	switch (result->depth) {
	  case 8:	p[c] = sum; break;
	  case 16:	((unsigned short *)p)[c] = sum; break;
	}
      }

  return 0;
#endif
}


#ifdef VIAM_OPENCV_SUPPORT

/* --- viam_geo_distort ---------------------------------------------- */

/** \brief Setup compensation of image distortion.
 *
 * \param[in] vh	Library handle
 * \param[in] image	Image used for initial acquisition
 * \param[in] source	Source image
 * \param[in] args	Filter arguments
 * \param[out] size	Computed final geometry
 * \param[out] depth	Computed final depth (unchanged)
 * \param[out] channels	Computed final number of channels (unchanged)
 *
 * \return 0 on success.
 */

int
viam_geo_distortsetup(viam_handle_t vh, viam_image_t image,
		      viam_filterargs_t *args, viam_geometry_t *geom, void **data)
{
  viam_cameracalibration_t cal;
  float *maps, *px, *py;
  CvMat mapx_, mapy_;
  CvMat i_, d_;
  int i, j, s;

  s = viam_geo_geosetup(vh, image, args, geom, data);
  if (s) return s;

  s = viam_image_calibration(vh, image, &cal, geom);
  if (s) return s;

  /* allocate and fill map */
  maps = malloc(geom->width*geom->height * 2 * sizeof(float));
  if (!maps) return viam_seterrno(vh, Eviamlib_MEMORY);
  mapx_ = cvMat(geom->height, geom->width, CV_32FC1, maps);
  mapy_ = cvMat(geom->height, geom->width, CV_32FC1,
		maps + geom->width*geom->height);

  i_ = cvMat(3, 3, CV_64FC1, cal.intrinsic);
  d_ = cvMat(1, 4, CV_64FC1, cal.distortion);
  cvInitUndistortMap(&i_, &d_, &mapx_, &mapy_);

  /* rescale maps depending on the source geometry (don't include crop
   * offset here: it is handled by a ROI on the source image). */
  px = maps; py = maps + geom->width*geom->height;
  for(i=0; i<geom->height; i++) for(j=0; j<geom->width; j++, px++, py++) {
    *px *= (args->right - args->left + 1) / args->towidth;
    *py *= (args->bottom - args->top + 1) / args->toheight;
  }

  /* store map */
  if (*data) free(*data);
  *data = maps;
  return 0;
}


/** \brief Compensate image distortion.
 *
 * \param[in] vh	Library handle
 * \param[in] image	Image used for initial acquisition
 * \param[in] source	Source image
 * \param[out] result	Resulting image
 * \param[in] args	Filter arguments
 *
 * \return 0 on success.
 */

int
viam_geo_remap(viam_handle_t vh, viam_image_t image, const IplImage *source,
	       IplImage *result, viam_filterargs_t *args, void *data)
{
  float *maps = data;
  CvRect crop;
  CvMat mapx_ = cvMat(result->height, result->width, CV_32FC1, maps);
  CvMat mapy_ = cvMat(result->height, result->width, CV_32FC1,
		      maps + result->width*result->height);
  int s = 0;

  crop.x = args->left;
  crop.y = args->top;
  crop.width = args->right - args->left + 1;
  crop.height = args->bottom - args->top + 1;
  cvSetImageROI((IplImage *)source, crop);
  cvRemap(source, result, &mapx_, &mapy_,
	  CV_INTER_LINEAR|CV_WARP_FILL_OUTLIERS, cvScalarAll(0));
  if (cvGetErrStatus()) s = Eviamlib_OPENCV;
  cvResetImageROI((IplImage *)source);

  return s;
}


/* --- viam_geo_rectify ---------------------------------------------- */

/** \brief Setup rectification of image (includes distortion compensation).
 *
 * \param[in] vh	Library handle
 * \param[in] image	Image used for initial acquisition
 * \param[in] source	Source image
 * \param[in] args	Filter arguments
 * \param[out] size	Computed final geometry
 * \param[out] depth	Computed final depth (unchanged)
 * \param[out] channels	Computed final number of channels (unchanged)
 *
 * \return 0 on success.
 */

int
viam_geo_rectifysetup(viam_handle_t vh, viam_image_t image,
		      viam_filterargs_t *args, viam_geometry_t *geom, void **data)
{
  viam_cameracalibration_t cal;
  float *maps, *px, *py;
  int i, j, s;

  s = viam_geo_geosetup(vh, image, args, geom, data);
  if (s) return s;

  s = viam_image_calibration(vh, image, &cal, geom);
  if (s) return s;

  /* allocate and fill map */
  maps = malloc(geom->width*geom->height * 2 * sizeof(float));
  if (!maps) return viam_seterrno(vh, Eviamlib_MEMORY);

  viam_initrectifmap(cal.intrinsic, cal.distortion, cal.rectification,
		     maps, maps + geom->width*geom->height,
		     geom->width, geom->height);

  /* rescale maps depending on the source geometry (don't include crop
   * offset here: it is handled by a ROI on the source image). */
  px = maps; py = maps + geom->width*geom->height;
  for(i=0; i<geom->height; i++) for(j=0; j<geom->width; j++, px++, py++) {
    *px *= (args->right - args->left + 1) / args->towidth;
    *py *= (args->bottom - args->top + 1) / args->toheight;
  }

  /* store map */
  if (*data) free(*data);
  *data = maps;
  return 0;
}

#endif /* VIAM_OPENCV_SUPPORT */


/* --- viam_geo_geosetup --------------------------------------------- */

/** \brief Compute resulting geometry for geometric filters.
 *
 * This function is used for all filters that include geometric reshaping
 * of the source image: VIAM_FILTER_SUBSAMPLE, VIAM_FILTER_DISTORT.
 *
 * \param[in] vh	Library handle
 * \param[in] source	Source image
 * \param[in] args	Filter arguments
 * \param[out] size	Computed final geometry
 * \param[out] depth	Computed final depth (unchanged)
 * \param[out] channels	Computed final number of channels (unchanged)
 *
 * \return 0 on success.
 */

int
viam_geo_geosetup(viam_handle_t vh, viam_image_t image,
		  viam_filterargs_t *args, viam_geometry_t *geom, void **data)
{
  /* check validity */
  if (args->left < 0 || args->top < 0 ||
      args->right < 0 || args->bottom < 0)
    return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  if (args->towidth < 0 || args->toheight < 0)
    return viam_seterrno(vh, Eviamlib_INVALID_VALUE);

  if (args->left >= args->right || args->top >= args->bottom)
    return viam_seterrno(vh, Eviamlib_INVALID_VALUE);

  /* change % into absolute values */
  if (args->left <= 1) args->left *= geom->width;
  if (args->right <= 1) args->right *= geom->width;
  if (args->top <= 1) args->top *= geom->height;
  if (args->bottom <= 1) args->bottom *= geom->height;
  if (args->towidth <= 1) args->towidth *= geom->width;
  if (args->toheight <= 1) args->toheight *= geom->height;

  /* clamp values to image bounds */
  if (args->left >= geom->width) args->left = geom->width - 1;
  if (args->top >= geom->height) args->top = geom->height - 1;
  if (args->right >= geom->width) args->right = geom->width - 1;
  if (args->bottom >= geom->height) args->bottom = geom->height -1;

  /* update absolute scaling factor and origin */
  geom->offx += args->left / geom->scalex;
  geom->offy += args->top / geom->scaley;

  geom->scalex *= args->towidth / (args->right - args->left + 1);
  geom->scaley *= args->toheight / (args->bottom - args->top + 1);

  /* compute size */
  geom->width = args->towidth;
  geom->height = args->toheight;

  /* depth and channels unchanged */
  return 0;
}
